import pytest
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from faker import Faker

from wbcore.contrib.authentication.models.users import User
from wbcore.contrib.permission.internal.registry import UserBackendRegistry

fake = Faker()


def create_internal_user():
    user = User.objects.create(is_active=True, username=fake.user_name(), email=fake.email(), is_internal=True)
    permission = Permission.objects.get_or_create(
        content_type=ContentType.objects.get_for_model(User), codename="is_internal_user"
    )[0]
    user.user_permissions.add(permission)
    return user


@pytest.mark.django_db
class TestBackend:
    def test_get_internal_users(self):
        user = create_internal_user()
        assert set(UserBackendRegistry().get_internal_users()) == {user}



# Generated by Django 5.2.9 on 2026-01-09 14:54

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('wbcore', '0014_biguserobjectpermission_system'),
    ]

    operations = [
        migrations.DeleteModel(
            name='GenericModel',
        ),
    ]



from typing import Iterable

from celery import shared_task
from django.conf import settings
from django.db import transaction
from django.dispatch import receiver
from django.utils.module_loading import import_string
from django.utils.translation import gettext
from rest_framework.reverse import reverse

from wbcore.contrib.authentication.models.users import User
from wbcore.shares.signals import handle_widget_sharing

from ...workers import Queue
from .models import NotificationType
from .models.notifications import Notification, send_notification_as_task


def get_user(u: User | int) -> User:
    if isinstance(u, User):
        return u
    elif isinstance(u, int):
        return User.objects.get(id=u)
    raise ValueError("Invalid user type")


@shared_task(queue=Queue.HIGH_PRIORITY.value)
def send_notification(
    code: str,
    title: str,
    body: str,
    user: User | Iterable[User | int] | int,
    reverse_name: str | None = None,
    reverse_args=None,
    reverse_kwargs=None,
    endpoint: str | None = None,
):
    """Method for dispatching a notification

    Args:
        code: The code pointing to a `NotificationType`
        title: The title of the notification
        body: The text of the notification
        user: The users that should receive the notification. Can either be a User object or an iterable of User objects
        reverse_name: The reverse name of an endpoint attached to this notification
        reverse_args: The arguments passed to the `reverse` function
        reverse_kwargs: The keyword arguments passed to the `reverse` function
        endpoint: The endpoint of resource. If provided, reverse_name + args + kwargs are disregarded
    """
    users = []
    if isinstance(user, list):
        for u in user:
            users.append(get_user(u))
    else:
        users.append(get_user(user))

    for user in users:
        notification_type = NotificationType.objects.get(code=code)
        if user.is_active:
            if not endpoint:
                endpoint = reverse(reverse_name, reverse_args, reverse_kwargs) if reverse_name else None
            notification = Notification.objects.create(
                title=title,
                body=body,
                user=user,
                notification_type=notification_type,
                endpoint=endpoint,
            )
            transaction.on_commit(
                lambda notification_pk=notification.pk: send_notification_as_task.delay(notification_pk)
            )


@receiver(handle_widget_sharing)
def share_notification(
    request, widget_relative_endpoint, share=False, share_message=None, share_recipients=None, **kwargs
):
    if share and share_recipients and share_message:
        for recipient in share_recipients:
            send_notification(
                code="workbench.system",
                title=gettext("{} shared a widget with you").format(
                    import_string(settings.WBCORE_DEFAULT_USER_NAME)(request.user)
                ),
                user=recipient,
                body=share_message,
                endpoint=widget_relative_endpoint,
            )



from datetime import timedelta
from unittest.mock import patch

import pytest
from bs4 import BeautifulSoup
from django.core import mail
from django.test.utils import override_settings

from wbcore.contrib.notifications.models import Notification
from wbcore.contrib.notifications.utils import base_domain


@pytest.fixture
def patched_import_string(mocker):
    return mocker.patch("wbcore.contrib.notifications.models.notifications.import_string")


@pytest.fixture
def mock_backend():
    from wbcore.contrib.notifications.backends.abstract_backend import (
        AbstractNotificationBackend,
    )

    class MockedBackend(AbstractNotificationBackend):
        @classmethod
        def send_notification(cls, notification):
            pass

        @classmethod
        def get_configuration(cls) -> dict:
            return {}

    return MockedBackend


@pytest.mark.django_db
class TestNotification:
    def test_factory(self, notification: Notification):
        assert isinstance(notification, Notification)
        assert notification.pk is not None

    def test_to_str(self, notification: Notification):
        assert str(notification) == f"{notification.user} {notification.title}"

    def test_endpoint_basename(self):
        assert Notification.get_endpoint_basename() == "wbcore:notifications:notification"

    def test_representation_value_key(self):
        assert Notification.get_representation_value_key() == "id"

    def test_representation_label_key(self):
        assert Notification.get_representation_label_key() == "{{title}}"

    def test_has_duplicated(self, notification_factory):
        notification = notification_factory.create()
        assert notification.has_duplicated() is False
        duplicated_notification = notification_factory.create(
            user=notification.user,
            title=notification.title,
            body=notification.body,
            endpoint=notification.endpoint,
            notification_type=notification.notification_type,
            created=notification.created + timedelta(minutes=60),
        )
        assert notification.has_duplicated(interval=61) is False
        assert notification.has_duplicated(interval=59) is False
        duplicated_notification.sent = duplicated_notification.created
        duplicated_notification.save()
        assert notification.has_duplicated(interval=59) is True

    @pytest.mark.parametrize("notification__endpoint", ["/wbcore/notifications/"])
    def test_full_valid_internal_endpoint(self, notification):
        assert notification.get_full_endpoint() == f"{base_domain()}{notification.endpoint}"

    @pytest.mark.parametrize("notification__endpoint", ["/some_invalid_namespace/notifications/"])
    def test_full_invalid_internal_endpoint(self, notification):
        assert notification.get_full_endpoint() is None

    @pytest.mark.parametrize("notification__endpoint", ["/wbcore/notifications/"])
    def test_full_internal_endpoint_as_shareable_link(self, notification):
        assert (
            notification.get_full_endpoint(as_shareable_internal_link=True)
            == f"{base_domain()}?widget_endpoint={notification.endpoint}"
        )

    @pytest.mark.parametrize("notification__endpoint", ["https://www.google.com"])
    def test_full_valid_external_endpoint(self, notification):
        assert notification.get_full_endpoint() == notification.endpoint

    @pytest.mark.parametrize("notification__endpoint", ["https.www.google.com"])
    def test_full_invalid_external_endpoint(self, notification):
        assert notification.get_full_endpoint() is None

    @patch.object(Notification, "_send_as_mail")
    @patch.object(Notification, "has_duplicated")
    def test_send_exclude_mail_with_duplicate(self, mock_has_duplicated, mock_send_as_mail, notification):
        mock_has_duplicated.return_value = False
        setting = notification.notification_type.get_setting_for_user(notification.user)
        setting.enable_email = True
        setting.save()

        assert mock_send_as_mail.call_count == 0
        notification.send()
        assert mock_send_as_mail.call_count == 1

        mock_has_duplicated.return_value = True
        notification.send()
        assert mock_send_as_mail.call_count == 1

    @patch.object(Notification, "_send_as_mail")
    def test_send_notification_task(
        self, patched_send_notification_email, notification, mock_backend, mocker, patched_import_string
    ):
        spy = mocker.spy(mock_backend, "send_notification")
        patched_import_string.return_value = mock_backend

        setting = notification.notification_type.get_setting_for_user(notification.user)
        setting.enable_email = False
        setting.save()

        notification.send()
        patched_send_notification_email.assert_not_called()

        setting.enable_email = True
        setting.save()
        notification.send()
        patched_send_notification_email.assert_called_once_with()

        setting.enable_web = True
        setting.save()
        notification.send()
        patched_import_string.assert_called_once()
        spy.assert_called_once_with(notification)

    @patch.object(Notification, "_send_as_mail")
    def test_send_notification_task_without_mail(
        self, patched_send_notification_email, notification, mock_backend, patched_import_string
    ):
        notification.user.wbnotification_user_settings.update(enable_email=False)
        patched_import_string.return_value = mock_backend

        notification.send()

        assert not patched_send_notification_email.called

    @override_settings(EMAIL_BACKEND="django.core.mail.backends.locmem.EmailBackend")
    @pytest.mark.parametrize("notification__endpoint", ["/wbcore/notifications/"])
    def test_send_notification_email(self, notification):
        setting = notification.notification_type.get_setting_for_user(notification.user)
        setting.enable_email = True
        setting.save()

        assert len(mail.outbox) == 0
        notification.send()
        assert len(mail.outbox) == 1
        soup = BeautifulSoup(mail.outbox[0].alternatives[0][0], "html.parser")
        assert soup.find("a", href=notification.get_full_endpoint(as_shareable_internal_link=True))



import urllib.parse
from collections import defaultdict
from typing import Any


def nest_row(row: dict[str, Any], sep="__") -> dict[str, Any]:
    """
    Utility function to nest row based on seperator.

    Examples:
        Row with key containing "__" such as "key__nested_key" will be converted to a dictionary "{"key": "nested_key"}"
    Args:
        row: The un-nested row
        sep: The seperated to use for nested separation

    Returns:
        A nested dictionary
    """
    res = defaultdict(dict)
    for k, v in row.items():
        splits = k.split(sep, maxsplit=1)
        if len(splits) == 2:
            res[splits[0]][splits[1]] = v
        else:
            res[k] = v
    return dict(res)


def parse_endpoint(request, segment: str, **extra_kwargs) -> str:
    parts = urllib.parse.urlsplit(request.get_full_path())
    new_path = f"{parts.path.rstrip('/')}/{segment}/" if parts.path else f"/{segment}/"
    res = urllib.parse.urlunsplit((parts.scheme, parts.netloc, new_path, parts.query, parts.fragment))
    return res


def get_import_export_identifier(view) -> str:
    try:
        identifier = view.IDENTIFIER
    except AttributeError:
        identifier = "{0.app_label}:{0.model}".format(view.get_content_type())
    return identifier



# Generated by Django 5.2.9 on 2026-01-12 12:23

from django.db import migrations, models
from tqdm import tqdm


def cleanup_duplicates(apps, schema_editor):
    EmailContact = apps.get_model("directory", "EmailContact")
    qs = EmailContact.objects.values("address").annotate(c=models.Count("*")).filter(c__gt=1).values("address")
    for rel in tqdm(qs, total=qs.count()):
        a = rel["address"]

        qs = EmailContact.objects.filter(address=a)
        # delete inactive contacts
        qs.filter(models.Q(entry__isnull=True) | models.Q(entry__is_active=False)).delete()

        # delete contact linked to company if it is already linked to person
        if qs.filter(entry__entry_type="Person").exists() and qs.filter(entry__entry_type="Company").exists():
            qs.filter(entry__entry_type="Company").delete()

        # for each duplicates, try to merge it, otherwise delete the email address
        if qs.count() > 1:
            first_contact = qs.earliest("entry_id")
            for tt in qs.exclude(id=first_contact.id):
                tt.delete()


class Migration(migrations.Migration):

    dependencies = [
        ('directory', '0014_alter_entry_relationship_managers_and_more'),
    ]

    operations = [
        migrations.RunSQL(sql="SET CONSTRAINTS ALL IMMEDIATE;"),
        migrations.RunPython(cleanup_duplicates),
        migrations.RunSQL(sql="SET CONSTRAINTS ALL DEFERRED;"),
        migrations.AlterField(
            model_name='emailcontact',
            name='address',
            field=models.EmailField(max_length=254, unique=True, verbose_name='Email Address'),
        ),
        migrations.AddConstraint(
            model_name='emailcontact',
            constraint=models.UniqueConstraint(condition=models.Q(('primary', True)), fields=('entry',), name='unique_primary_entry_address'),
        ),
    ]



from django.db.models.signals import ModelSignal

# Signal fired before merged/obsolete object are deleted during a merge phase
pre_merge = ModelSignal(use_caching=False)
post_merge = ModelSignal(use_caching=False)



from django.conf import settings
from django.contrib.auth.models import Group
from django.db.models import Q, QuerySet
from django.utils.module_loading import import_string

from wbcore.contrib.authentication.models import User


class UserBackendRegistry:
    def __init__(self):
        internal_users_backend_path = getattr(
            settings, "USER_BACKEND", "wbcore.contrib.permission.internal.backend.UserBackend"
        )
        internal_users_backend_class = import_string(internal_users_backend_path)
        self.backend = internal_users_backend_class()

    def get_internal_groups(self) -> QuerySet[Group]:
        return self.backend.get_internal_groups()

    def get_internal_users(self) -> QuerySet[User]:
        return User.objects.filter(Q(is_internal=True) | Q(id__in=self.backend.get_internal_users().values("id")))

    def refresh_users(self, reset_all: bool = False):
        if reset_all:
            User.objects.filter(is_internal=True).update(is_internal=False)
        users = []
        for user in self.get_internal_users().filter(is_active=True):
            user.is_internal = True
            users.append(user)
        User.objects.bulk_update(users, ["is_internal"])


def get_internal_groups() -> QuerySet[Group]:
    """
    Return the cached groups of internals users

    Returns:
        A queryset of group corresponding to the internal notion defined by the set UserBackend

    Raises:
        ValueError: If user backend path does not correspond to a valid module
    """
    return UserBackendRegistry().internal_groups



# Generated by Django 5.2.9 on 2026-01-28 14:41

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='GroupObjectPermission',
            fields=[
                ('object_pk', models.CharField(max_length=255, verbose_name='object ID')),
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False, unique=True)),
                ('editable', models.BooleanField(default=True)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='auth.group')),
                ('permission', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='auth.permission')),
            ],
            options={
                'abstract': False,
                'default_related_name': 'groupobjectpermissions',
                'indexes': [models.Index(fields=['content_type', 'object_pk'], name='wbcore_perm_content_7e5298_idx'), models.Index(fields=['content_type', 'object_pk', 'group'], name='wbcore_perm_content_e0e7e3_idx')],
                'unique_together': {('group', 'permission', 'object_pk')},
            },
        ),
        migrations.CreateModel(
            name='UserObjectPermission',
            fields=[
                ('object_pk', models.CharField(max_length=255, verbose_name='object ID')),
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False, unique=True)),
                ('editable', models.BooleanField(default=True)),
                ('system', models.BooleanField(default=False)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
                ('permission', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='auth.permission')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
                'default_related_name': 'userobjectpermissions',
                'indexes': [models.Index(fields=['content_type', 'object_pk'], name='wbcore_perm_content_753f13_idx'), models.Index(fields=['content_type', 'object_pk', 'user'], name='wbcore_perm_content_9efa9b_idx')],
                'unique_together': {('user', 'permission', 'object_pk')},
            },
        ),
    ]



from celery import shared_task
from tqdm import tqdm

from wbcore.contrib.permission.models.mixins import PermissionObjectModelMixin
from wbcore.utils.itertools import get_inheriting_subclasses
from wbcore.workers import Queue


@shared_task(queue=Queue.EXTENDED_BACKGROUND.value)
def reload_permissions_as_task(
    prune_existing: bool | None = True, force_pruning: bool | None = False, debug: bool = False
):
    for subclass in get_inheriting_subclasses(PermissionObjectModelMixin):
        gen = subclass.objects.iterator()
        if debug:
            gen = tqdm(gen, total=subclass.objects.count())
        for instance in gen:
            instance.reload_permissions(prune_existing=prune_existing, force_pruning=force_pruning)



from django.urls import include, path

from wbcore.contrib.content_type.viewsets import ContentTypeRepresentationViewSet, DynamicObjectIDRepresentationViewSet
from wbcore.routers import WBCoreRouter

router = WBCoreRouter()

router.register(r"contenttyperepresentation", ContentTypeRepresentationViewSet, basename="contenttyperepresentation")
router.register(
    r"dynamiccontenttyperepresentation",
    DynamicObjectIDRepresentationViewSet,
    basename="dynamiccontenttyperepresentation",
)

urlpatterns = [path("", include(router.urls))]



from django.utils.translation import gettext as _

from wbcore.menus import ItemPermission, MenuItem

DATA_MENUITEM = MenuItem(
    label=_("Data"),
    endpoint="wbcore:workflow:data-list",
    permission=ItemPermission(method=lambda request: request.user.is_internal, permissions=["workflow.view_data"]),
    add=MenuItem(
        label=_("Create Data"),
        endpoint="wbcore:workflow:data-list",
        permission=ItemPermission(method=lambda request: request.user.is_internal, permissions=["workflow.add_data"]),
    ),
)



import importlib
import re
from contextlib import suppress
from functools import cached_property

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.dispatch import receiver
from django.shortcuts import get_object_or_404
from django.urls import resolve
from django.urls.exceptions import NoReverseMatch
from django.utils.translation import gettext as _
from rest_framework import status
from rest_framework.decorators import action, api_view, permission_classes
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.reverse import reverse

from wbcore import serializers as wb_serializer
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.display.instance_display import Display
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)
from wbcore.signals.instance_buttons import add_extra_button
from wbcore.utils.models import MergeMixin as ModelMergeMixin
from wbcore.utils.models import merge_as_task


def parse_query_parameters_list(list_str):
    try:
        return list(set(map(lambda x: int(x), filter(None, list_str.split(",")))))
    except ValueError:
        return ""


def set_identifier(identifier):
    def decorator(func):
        func.identifier = identifier
        return func

    return decorator


@api_view(["GET"])
@permission_classes([IsAuthenticated])
def api_endpoints_root(request, format=None):
    try:
        endpoints = dict()
        for wb_endpoint in settings.WB_ENDPOINTS:
            try:
                endpoints[wb_endpoint] = reverse(f"{wb_endpoint}:api-root", request=request, format=format)
            except NoReverseMatch:
                pass

        return Response(endpoints)

    except AttributeError:
        return Response({"error": "No Endpoints specified."}, status=status.HTTP_400_BAD_REQUEST)


class CloneMixin:
    """
    This Mixin can be used to inherit the clone functionality out of the box. This mixin will register the additional resources
    and the appropriate button to allow user to clone a given instance
    """

    def get_clone_button_serializer_class(self, instance):
        return self.get_serializer_class()

    def get_clone_button_instance_display(self) -> Display | None:
        """
        Return the instance display to show in the clone action button. Default to return None (i.e. No form)
        """
        return None

    @action(methods=["PATCH"], detail=True)
    def clone(self, request, pk=None):
        instance = get_object_or_404(self.queryset.model, pk=pk)
        if (clone_fct := getattr(instance, "clone", None)) and callable(clone_fct):
            kwargs = {}
            if request.data:
                serializer = self.get_clone_button_serializer_class(instance)(data=request.data)
                serializer.is_valid(raise_exception=False)
                kwargs = serializer.data
            cloned_obj = clone_fct(user=request.user, **kwargs)
            if (
                cloned_obj
                and (clone_id := getattr(cloned_obj, "id", None))
                and (endpoint_config_class := getattr(self, "endpoint_config_class", None))
            ):
                endpoint_config = endpoint_config_class(self, request, cloned_obj)
                if endpoint := endpoint_config.get_instance_endpoint():
                    return Response(
                        {
                            "endpoint": re.sub(
                                pattern=r"[^\/]+(?=\/$|$)",
                                repl=str(clone_id),
                                string=endpoint,
                            )
                        }
                    )
        return Response({})


@receiver(add_extra_button)
def add_clone_extra_button(sender, instance, request, view, pk=None, **kwargs):
    if instance and pk and issubclass(view.__class__, CloneMixin):
        object = view.get_object()
        if getattr(object, "is_cloneable", True) and (endpoint_basename := object.get_endpoint_basename()):
            endpoint = reverse(f"{endpoint_basename}-clone", args=[pk], request=request)
            identifier = getattr(view, "IDENTIFIER", "{0.app_label}:{0.model}".format(view.get_content_type()))
            return bt.ActionButton(
                method=RequestType.PATCH,
                identifiers=(identifier,),
                endpoint=endpoint,
                action_label=_("Clone"),
                title=_("Clone"),
                label=_("Clone"),
                description_fields=_('<strong>Clone {}</strong> "{}"').format(object._meta.verbose_name, object),
                icon=WBIcon.COPY.icon,
                serializer=view.get_clone_button_serializer_class(object),
                instance_display=view.get_clone_button_instance_display(),
            )


class MergeMixin:
    """
    This Mixin can be used to inherit the clone functionality out of the box. This mixin will register the additional resources
    and the appropriate button to allow user to merge a given instance
    """

    @cached_property
    def model(self):
        return self.queryset.model

    def has_merge_permission(self, user) -> bool:
        return user.has_perm(f"administrate_{self.model._meta.model_name}")

    def get_merged_object_representation_serializer(self):
        """
        Return the representation serializer used for the action button merged object field. Default to the model representation endpoint serializer
        """
        view_func = resolve(reverse(self.model.get_representation_endpoint())).func
        module = importlib.import_module(view_func.__module__)
        view = getattr(module, view_func.__name__)
        return view.serializer_class

    def get_merged_object_representation_serializer_filter_params(self) -> dict[str, str]:
        """
        Return the representation serializer filter kwargs. Default to an empty dict
        """
        return dict()

    def add_extra_buttons(self, request):
        if self.has_merge_permission(request.user) and (pk := self.kwargs.get("pk")):
            object = self.get_object()
            if (
                isinstance(object, ModelMergeMixin)
                and object.is_mergeable
                and (endpoint_basename := object.get_endpoint_basename())
            ):
                endpoint = reverse(f"{endpoint_basename}-merge", args=[pk], request=request)

                class MergeSerializer(wb_serializer.ModelSerializer):
                    merged_object = wb_serializer.PrimaryKeyRelatedField(
                        queryset=self.model.objects.all(), label="Merged with"
                    )
                    _merged_object = self.get_merged_object_representation_serializer()(
                        source="merged_object",
                        filter_params=self.get_merged_object_representation_serializer_filter_params(),
                    )

                    class Meta:
                        model = self.model
                        fields = [
                            "id",
                            "merged_object",
                            "_merged_object",
                        ]

                return {
                    bt.ActionButton(
                        method=RequestType.PATCH,
                        endpoint=endpoint,
                        action_label=_("Merge"),
                        title=_("Merge"),
                        label=_("Merge"),
                        icon=WBIcon.MERGE.icon,
                        serializer=MergeSerializer,
                        instance_display=create_simple_display([["merged_object"]]),
                    ),
                }

    @action(methods=["PATCH"], detail=True)
    def merge(self, request, pk=None):
        if self.has_merge_permission(request.user):
            if pk and hasattr(self.model, "merge"):
                instance = get_object_or_404(self.queryset.model, pk=pk)
                merged_instance = get_object_or_404(self.queryset.model, pk=request.POST.get("merged_object", None))
                with suppress(ContentType.DoesNotExist):
                    content_type = ContentType.objects.get_for_model(self.model)
                    merge_as_task.delay(content_type.id, instance.id, merged_instance.id)
                    return Response({"status": "Merged ongoing"})
            return Response({}, status=400)
        return Response({"_detail": "User does not have the permission to merge"}, status=status.HTTP_403_FORBIDDEN)



# Generated by Django 5.2.9 on 2026-02-02 11:27
import hashlib

from django.conf import settings
from django.db import migrations, models
from tqdm import tqdm

from wbcore.contrib.notifications.utils import get_checksum


def migrate_checksum(apps, schema_editor):
    Notification = apps.get_model('notifications', 'Notification')
    qs = Notification.objects.all()
    objs = []
    for notification in tqdm(qs, total=qs.count()):
        notification.checksum = get_checksum(notification.title, notification.body, notification.endpoint)
        objs.append(notification)
    Notification.objects.bulk_update(objs, ["checksum"], batch_size=1000)

class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0009_alter_notificationtypesetting_options_and_more'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name='notification',
            name='checksum',
            field=models.CharField(null=True, blank=True, max_length=64),
            preserve_default=False,
        ),
        migrations.RunPython(migrate_checksum),
        migrations.AlterField(
            model_name='notification',
            name='body',
            field=models.TextField(default=''),
        ),
        migrations.AlterField(
            model_name='notification',
            name='checksum',
            field=models.CharField(default=None, max_length=64),
            preserve_default=False,
        ),
        migrations.AddIndex(
            model_name='notification',
            index=models.Index(fields=['user', 'notification_type', 'checksum'], name='notificatio_user_id_b90c58_idx'),
        ),
    ]



# Generated by Django 5.2.9 on 2026-01-26 12:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('authentication', '0007_alter_token_unique_together_token_unique_token'),
    ]

    operations = [
        migrations.AddField(
            model_name='user',
            name='is_internal',
            field=models.BooleanField(default=False, help_text='If true, define this user as internal which grants specific permissions.', verbose_name='Internal'),
        ),
    ]



from django.apps import AppConfig
from django.db.models.signals import post_migrate


class PermissionAppConfig(AppConfig):
    name = "wbcore.contrib.permission"
    label = "wbcore_permission"

    def ready(self):
        from wbcore.contrib.permission.management import refresh_internal_users

        post_migrate.connect(
            refresh_internal_users,
            dispatch_uid="permission.refresh_internal_users",
        )



import urllib
from contextlib import suppress
from datetime import timedelta

from celery import shared_task
from django.conf import settings
from django.core.mail import EmailMultiAlternatives
from django.core.validators import URLValidator, ValidationError
from django.db import models
from django.template.loader import get_template
from django.urls import Resolver404, resolve
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.html import strip_tags
from django.utils.module_loading import import_string

from wbcore.contrib.authentication.models.users import User
from wbcore.contrib.notifications.utils import base_domain, create_notification_type, get_checksum
from wbcore.workers import Queue


class Notification(models.Model):
    title = models.CharField(max_length=255)
    body = models.TextField(default="")
    endpoint = models.CharField(max_length=2048, null=True, blank=True)

    user = models.ForeignKey(to=User, related_name="notifications_notifications", on_delete=models.CASCADE)
    notification_type = models.ForeignKey(
        to="notifications.NotificationType", related_name="notifications", on_delete=models.CASCADE
    )

    created = models.DateTimeField(auto_now_add=True)
    sent = models.DateTimeField(null=True, blank=True)
    read = models.DateTimeField(null=True, blank=True)
    checksum = models.CharField(max_length=64)

    def __str__(self) -> str:
        return f"{self.user} {self.title}"

    class Meta:
        verbose_name = "Notification"
        verbose_name_plural = "Notifications"

        notification_types = [
            create_notification_type(
                "workbench.system",
                "System Notifications",
                "System Notifications.",
                True,
                True,
                False,
            ),
        ]
        indexes = [
            models.Index(fields=["user", "notification_type", "checksum"]),
        ]
        # constraints = [
        #     models.UniqueConstraint(fields=["user", "checksum"], name="checksum_unique"),
        # ]

    @cached_property
    def is_endpoint_internal(self) -> bool:
        with suppress(Resolver404):
            if self.endpoint:
                resolve(
                    urllib.parse.urlsplit(urllib.parse.unquote(self.endpoint)).path
                )  # we need to truncate query parameters
                return True
        return False

    @cached_property
    def is_endpoint_valid(self) -> bool:
        try:
            URLValidator()(self.endpoint)
            return True
        except ValidationError:
            return False

    def _send_as_mail(self):
        """Sends out a notification to the user specified inside the notification"""

        context = {
            "title": self.title,
            "message": self.body or "",
            "notification_share_url": self.get_full_endpoint(as_shareable_internal_link=True),
            "notification_endpoint": self.get_full_endpoint(),
        }
        rendered_template = get_template("notifications/notification_template.html").render(context)
        msg = EmailMultiAlternatives(
            subject=self.title,
            body=strip_tags(rendered_template),
            from_email=getattr(settings, "WBCORE_NOTIFICATION_EMAIL_FROM", "no_reply@stainly.com"),
            to=[self.user.email],  # type: ignore
        )
        msg.attach_alternative(rendered_template, "text/html")
        msg.send()

    def save(self, *args, **kwargs):
        if not self.checksum:
            self.checksum = self.get_checksum()
        super().save(*args, **kwargs)

    def has_duplicated(self, interval: int = 60 * 4) -> bool:
        return (
            Notification.objects.exclude(id=self.id)
            .filter(
                user=self.user,
                notification_type=self.notification_type,
                checksum=self.checksum,
                sent__isnull=False,
                created__gt=self.created - timedelta(minutes=interval),
                created__lt=self.created + timedelta(minutes=interval),
            )
            .exists()
        )

    def send(self):
        notification_user_setting = self.notification_type.get_setting_for_user(self.user)

        # we do not sent notification through email if we detect similar already sent notification
        if notification_user_setting.enable_email and not self.has_duplicated():
            self._send_as_mail()
        if notification_user_setting.enable_web or notification_user_setting.enable_mobile:
            backend = import_string(settings.NOTIFICATION_BACKEND)
            backend.send_notification(self)

        # mark this notification as sent
        self.sent = timezone.now()
        self.save()

    def get_checksum(self) -> str:
        return get_checksum(self.title, self.body, self.endpoint)

    def get_full_endpoint(self, as_shareable_internal_link: bool = False) -> str | None:
        if self.is_endpoint_internal:
            if as_shareable_internal_link:
                return f"{base_domain()}?widget_endpoint={self.endpoint}"
            else:
                return f"{base_domain()}{self.endpoint}"
        elif self.is_endpoint_valid:
            return self.endpoint

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcore:notifications:notification"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{title}}"


@shared_task(queue=Queue.HIGH_PRIORITY.value)
def send_notification_as_task(notification_pk: int):
    """A celery task to send out a notification via email, web or mobile

    Args:
        notification_pk: The primary key of the notification that is going to be send out
    """

    notification = Notification.objects.get(pk=notification_pk)
    notification.send()
